from coopgraph.grids import RectGrid
from coopgraph.dataStructs import GridPoint
from coopstructs.geometry import Rectangle
import pygame
from coopgame.colors import Color
from typing import Dict, Callable
from coopgame.enums import GridDrawType
import coopgame.pygamehelpers as help
from typing import List
from coopstructs.vectors import Vector2
import numpy as np
from coopbugger.monitoredclass import MonitoredClass

class PyGridHandler(MonitoredClass):

    def __init__(self):
        super().__init__()
        self.hover_grid_pos = None

    @MonitoredClass.timer
    def _check_and_handle_hover_grid_change(self, old_hover: Vector2, new_hover: Vector2, handlers: List[Callable[[Vector2, Vector2], None]]):
        if new_hover != self.hover_grid_pos:
            [handler(old_hover, new_hover) for handler in handlers or []]

    @MonitoredClass.timer
    def handle_hover_over(self,
                          grid: RectGrid,
                          area_rect: Rectangle,
                          on_hover_handlers: List[Callable[[Vector2], None]] = None,
                          on_hover_changed_handlers: List[Callable[[Vector2, Vector2], None]] = None,
                          draw_scale_matrix = None):
        # get hovered grid pos
        mouse_pos = help.mouse_pos_as_vector()
        new_hover = self.get_mouse_grid_pos(area_rect, mouse_pos, grid.grid_from_coord, draw_scale_matrix)

        # check and handle when hover grid pos changed
        self._check_and_handle_hover_grid_change(old_hover=self.hover_grid_pos, new_hover=new_hover, handlers=on_hover_changed_handlers)

        # Update the grid pos
        self.hover_grid_pos = new_hover

        # handle grid hover
        [handler(self.hover_grid_pos) for handler in on_hover_handlers or []]

    @MonitoredClass.timer
    def grid_box_rectangle(self, width, height, grid:RectGrid, margin: int = 0) -> Rectangle:

        grid_box_height = height / grid.nRows - 2 * margin
        grid_box_width = width / grid.nColumns - 2 * margin

        return Rectangle(x=0, y=0, height=grid_box_height, width=grid_box_width)

    @MonitoredClass.timer
    def draw_overlay_to_surface(self
             , surface:pygame.Surface
             , grid: RectGrid
             , margin: int = 0
             , highlight_grid_cells: Dict[Vector2, Color] = None
             , outlined_grid_cells: Dict[Vector2, Color] = None
             , draw_scale_matrix = None
                                ):
        grid_box_rect = self.grid_box_rectangle(width=surface.get_width(), height=surface.get_height(), grid=grid, margin=margin)
        self._draw_highlighted_grids(surface=surface
                                    , grid_box_rect=grid_box_rect
                                    , highlight_grid_cells=highlight_grid_cells
                                    , draw_scale_matrix = draw_scale_matrix)
        self._draw_outlined_grids(surface=surface
                                    , grid_box_rect=grid_box_rect
                                    , outlined_grid_cells=outlined_grid_cells
                                    , margin=margin
                                    , draw_scale_matrix = draw_scale_matrix)

    @MonitoredClass.timer
    def draw_base_to_surface(self
             , surface:pygame.Surface
             , grid: RectGrid
             , grid_draw_type: GridDrawType = None
             , grid_color: Color = None
             , margin:int=0
             , draw_scale_matrix: np.array = None
             , center_guide_color: Color = None):
        if grid_draw_type is None:
            grid_draw_type = GridDrawType.BOXES

        if grid_color is None:
            grid_color = Color.BLACK

        if grid_draw_type == GridDrawType.BOXES:
            self._draw_grid_boxes(surface=surface, grid=grid, margin=margin, grid_color=grid_color, draw_scale_matrix = draw_scale_matrix)
        elif grid_draw_type == GridDrawType.LINES:
            self._draw_grid_lines(surface=surface, grid=grid, margin=margin, grid_color=grid_color, draw_scale_matrix = draw_scale_matrix)

        if center_guide_color is not None:
            rs, cs = self.define_center_grids(grid)
            highlights = {}
            row_highlights = {Vector2(r, c): center_guide_color for r in rs for c in range(0, grid.nColumns) }
            col_highlights = {Vector2(r, c): center_guide_color for r in range(0, grid.nRows) for c in cs}
            grid_box_rect = self.grid_box_rectangle(surface.get_width(), surface.get_height(), grid, margin)
            highlights.update(col_highlights)
            highlights.update(row_highlights)
            self._draw_highlighted_grids(surface,
                                         grid_box_rect=grid_box_rect,
                                         highlight_grid_cells=highlights,
                                         draw_scale_matrix=draw_scale_matrix)

    @MonitoredClass.timer
    def define_center_grids(self, grid: RectGrid):
        if grid.nRows % 2 == 0:
            center_rows = [grid.nRows / 2 + 1, grid.nRows / 2]
        else:
            center_rows = [(grid.nRows - 1 / 2) + 1]

        if grid.nColumns % 2 == 0:
            center_cols = [grid.nColumns / 2 + 1, grid.nColumns / 2]
        else:
            center_cols = [(grid.nColumns - 1 / 2) + 1]

        return (center_rows, center_cols)

    @MonitoredClass.timer
    def _draw_outlined_grids(self, surface:pygame.Surface, grid_box_rect: Rectangle, margin=0, outlined_grid_cells: Dict[Vector2, Color] = None, draw_scale_matrix = None):
        if outlined_grid_cells is None or len(outlined_grid_cells) == 0:
            return

        for grid_pos, color in outlined_grid_cells.items():
            if color is None:
                color = Color.YELLOW

            rect = Rectangle(x=(margin + grid_box_rect.width) * grid_pos.x + margin
                             , y=(margin + grid_box_rect.height) * grid_pos.y + margin
                             , height=grid_box_rect.height
                             , width=grid_box_rect.width)
            my_image = pygame.Surface((rect.width, rect.height), pygame.SRCALPHA)
            pygame.draw.rect(my_image, color.value, my_image.get_rect(), 3)
            surface.blit(my_image, (rect.x, rect.y))

    @MonitoredClass.timer
    def _draw_highlighted_grids(self,
                                surface:pygame.Surface,
                                grid_box_rect: Rectangle,
                                highlight_grid_cells: Dict[Vector2, Color] = None,
                                draw_scale_matrix = None):
        if highlight_grid_cells is None or len(highlight_grid_cells) == 0:
            return

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        for grid_pos, color in highlight_grid_cells.items():
            if color is None:
                color = Color.YELLOW
            points = help.scaled_points_of_a_rect(grid_box_rect,
                                                  grid_pos,
                                                  draw_scale_matrix)[:, :2]

            help.draw_polygon(surface, list(points), color)

    @MonitoredClass.timer
    def _draw_grid_boxes(self, surface:pygame.Surface, grid: RectGrid, margin=0, grid_color: Color = None, draw_scale_matrix = None):
        grid_box_rect = self.grid_box_rectangle(surface.get_width(), surface.get_height(), grid, margin)

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)


        for y in range(0, grid.nRows):
            for x in range(0, grid.nColumns):
                points = help.scaled_points_of_a_rect(grid_box_rect, Vector2(x, y), draw_scale_matrix)[:, :2]
                help.draw_polygon(surface, list(points), grid_color)

    @MonitoredClass.timer
    def _draw_grid_lines(self, surface:pygame.Surface, grid: RectGrid, margin=0, grid_color: Color = None, draw_scale_matrix = None):
        grid_box_rect = self.grid_box_rectangle(surface.get_width(), surface.get_height(), grid, margin)

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        '''Draw each line'''
        for x in range(0, grid.nColumns + 1):
            point_x1 = x * (grid_box_rect.width + margin)
            point_y1 = 0
            start = draw_scale_matrix.dot(np.array([point_x1, point_y1, 0, 1]))
            point_x2 = x * (grid_box_rect.width + margin)
            point_y2 = (grid_box_rect.height + margin) * grid.nRows
            end = draw_scale_matrix.dot(np.array([point_x2, point_y2, 0, 1]))
            _s = tuple(start[:2].astype(int))
            _e = tuple(end[:2].astype(int))
            pygame.draw.aaline(surface, grid_color.value, _s, _e)
        for y in range(0, grid.nRows + 1):
            start = draw_scale_matrix.dot(np.array([0, y * (grid_box_rect.height + margin), 0, 1]))
            end = draw_scale_matrix.dot(np.array([(grid_box_rect.width + margin) * grid.nColumns, y * (grid_box_rect.height + margin), 0, 1]))
            _s = tuple(start[:2])
            _e = tuple(end[:2])
            pygame.draw.aaline(surface, grid_color.value, _s, _e)

    @MonitoredClass.timer
    def get_mouse_grid_pos(self, game_area_rect: Rectangle, mouse_pos: Vector2, coord_to_grid_converter: Callable[[Vector2, Rectangle], Vector2], draw_scale_matrix = None):
        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        draw_scale_matrix_inv = np.linalg.inv(draw_scale_matrix)

        # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
        """Gets the mouse position and converts it to a grid position"""
        mouse_game_area_coord = help.game_area_coords_from_parent_coords(parent_coords=mouse_pos, game_area_surface_rectangle=game_area_rect)
        mouse_over_grid_coords = help.viewport_point_on_plane(mouse_game_area_coord, game_area_rect, draw_scale_matrix, margin=0)
        points = np.array([mouse_over_grid_coords[0], mouse_over_grid_coords[1], mouse_over_grid_coords[2], 1])
        grid_coords = draw_scale_matrix_inv.dot(np.transpose(points))

        # return coord_to_grid_converter(mouse_game_area_coord, game_area_rect)
        return coord_to_grid_converter(Vector2(grid_coords[0], grid_coords[1]), game_area_rect)

    @MonitoredClass.timer
    def grid_point_to_viewport_point(self, surface: pygame.Surface, grid, grid_pos: Vector2, grid_point_type: GridPoint = GridPoint.CENTER, margin: int = 0, draw_scale_matrix = None):
        grid_box_rect = self.grid_box_rectangle(surface.get_width(), surface.get_height(), grid, margin)
        grid_box_scaled_points = help.scaled_points_of_a_rect(grid_box_rect, grid_pos=grid_pos, draw_scale_matrix=draw_scale_matrix)

        if grid_point_type == GridPoint.CENTER:
            x = sum(point[0] for point in grid_box_scaled_points)/len(grid_box_scaled_points)
            y = sum(point[1] for point in grid_box_scaled_points) / len(grid_box_scaled_points)
            return Vector2(x, y)
        elif grid_point_type == GridPoint.ORIGIN:
            return Vector2(grid_box_scaled_points[0][0], grid_box_scaled_points[0][1])
        else:
            raise NotImplementedError(f"Unimplemented grid_point_type {grid_point_type}")

    @MonitoredClass.timer
    def grid_pos_definition(self,
                            grid: RectGrid,
                            grid_pos: Vector2,
                            game_area_rect: Rectangle) -> Rectangle:
        grid_box_rect = self.grid_box_rectangle(game_area_rect.width, game_area_rect.height, grid)
        origin = Vector2(grid_pos.x * grid_box_rect.width, grid_pos.y * grid_box_rect.height)

        return Rectangle(origin.x, origin.y, grid_box_rect.height, grid_box_rect.width)



if __name__ == "__main__":
    grid = RectGrid(10, 10)
    grid_pos = Vector2(5, 9)
    game_area_rect = Rectangle(0, 0, 1500, 2000)

    grid_handler = PyGridHandler()
    grid_def = grid_handler.grid_pos_definition(grid, grid_pos, game_area_rect)

    print(grid_def)
    print(grid_def.center)




from coopgame.colors import Color
import pygame
from coopgraph.graphs import Graph
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle
import coopgame.pygamehelpers as help
from typing import Callable, Tuple


class PyGraphHandler:
    def _draw_graph_edges(self,
                          surface: pygame.Surface,
                          graph: Graph,
                          coordinate_converter: Callable[[Vector2], Vector2] = None,
                          enabled_color: Color = None,
                          disabled_color: Color = None,
                          draw_scale_matrix=None,
                          draw_label_color: Color = None,
                          draw_label_font: pygame.font.Font = None,
                          label_offset: Tuple[int, int] = None
                          ):

        if not coordinate_converter:
            coordinate_converter = lambda x: x

        if not enabled_color:
            enabled_color = Color.BLUE

        if not disabled_color:
            disabled_color = Color.SLATE_GRAY

        if draw_label_color:
            draw_label_font = draw_label_font if draw_label_font is not None else pygame.font.Font(None, 25)
            label_offset = label_offset if label_offset is not None else (10, 10)

        for edge in graph.edges:
            start = coordinate_converter(edge.start.pos)
            end = coordinate_converter(edge.end.pos)

            width = 3 if graph.edge_between(edge.end, edge.start) is not None else 1

            scaled_pos = help.scaled_points_of_points([start, end], draw_scale_matrix=draw_scale_matrix)
            color = enabled_color if edge.enabled() else disabled_color
            pygame.draw.line(surface, color.value, (scaled_pos[0][0], scaled_pos[0][1]),
                             (scaled_pos[1][0], scaled_pos[1][1]), width)

            if draw_label_color is not None:
                mid = ((scaled_pos[0][0] + scaled_pos[1][0]) / 2, (scaled_pos[0][1] + scaled_pos[1][1]) / 2)
                help.draw_text(edge.id, surface, draw_label_font,
                               offset_rect=Rectangle(mid[0] + label_offset[0], mid[1] + label_offset[1], 100, 100),
                               color=draw_label_color)

    def _draw_graph_nodes(self,
                          surface: pygame.Surface,
                          graph: Graph,
                          coordinate_converter: Callable[[Vector2], Vector2] = None,
                          color: Color = None,
                          radius: int = 1,
                          width: int = 0,
                          draw_scale_matrix=None,
                          draw_label_color: Color = None,
                          draw_label_font: pygame.font.Font = None,
                          label_offset: Tuple[int, int] = None):
        if not coordinate_converter:
            coordinate_converter = lambda x: x

        if not color:
            color = Color.ORANGE

        txt_dict = {}
        for node in graph.nodes:
            position = coordinate_converter(node.pos)
            scaled_pos = help.scaled_points([position], draw_scale_matrix=draw_scale_matrix)
            position = (int(scaled_pos[0].x), int(scaled_pos[0].y))

            pygame.draw.circle(surface, color.value, position, radius, width)

            txt_dict.setdefault(scaled_pos[0], []).append(node.name)

        if draw_label_color is not None:
            draw_label_font = draw_label_font if draw_label_font is not None else pygame.font.Font(None, 25)
            label_offset = label_offset if label_offset is not None else (10, 10)

            for pos, labels in txt_dict.items():
                help.draw_text(",".join(labels), surface, draw_label_font,
                               offset_rect=Rectangle(pos.x + label_offset[0], pos.y + label_offset[1], 100, 100),
                               color=draw_label_color)

    @staticmethod
    def draw_articulation_points(surface: pygame.Surface,
                                 graph: Graph,
                                 color: Color = None,
                                 draw_scale_matrix=None
                                 ):
        if graph is None:
            return

        articulation_points = graph.AP()

        if color is None:
            color = Color.ORANGE

        scaled_pos = help.scaled_points([node.pos for node in articulation_points.keys()],
                                        draw_scale_matrix=draw_scale_matrix)

        for point in scaled_pos:
            pygame.draw.circle(surface, color.value, (int(point.x), int(point.y)), 10, 1)

    def draw_to_surface(self, surface: pygame.Surface,
                        graph: Graph,
                        coordinate_converter: Callable[[Vector2], Vector2] = None,
                        node_color: Color = None,
                        enabled_edge_color: Color = None,
                        disabled_edge_color: Color = None,
                        draw_scale_matrix=None,
                        node_label_color: Color = None,
                        edge_label_color: Color = None,
                        draw_label_font: pygame.font.Font = None,
                        draw_articulation_points: bool = False):
        if not coordinate_converter:
            coordinate_converter = lambda x: x

        self._draw_graph_edges(surface,
                               graph=graph,
                               coordinate_converter=coordinate_converter,
                               enabled_color=enabled_edge_color,
                               disabled_color=disabled_edge_color,
                               draw_scale_matrix=draw_scale_matrix,
                               draw_label_color=edge_label_color,
                               draw_label_font=draw_label_font)
        self._draw_graph_nodes(surface,
                               graph=graph,
                               coordinate_converter=coordinate_converter,
                               color=node_color,
                               draw_scale_matrix=draw_scale_matrix,
                               draw_label_color=node_label_color,
                               draw_label_font=draw_label_font)

        if draw_articulation_points:
            self.draw_articulation_points(surface,
                                          graph,
                                          Color.ORANGE,
                                          draw_scale_matrix=draw_scale_matrix)


import pygame

def pygame_key_mapper(val: int):
    mapper = {
        pygame.K_BACKSPACE: "K_BACKSPACE",
        pygame.K_TAB: "K_TAB",
        pygame.K_CLEAR: "K_CLEAR",
        pygame.K_RETURN: "K_RETURN",
        pygame.K_PAUSE: "K_PAUSE",
        pygame.K_ESCAPE: "K_ESCAPE",
        pygame.K_SPACE: "K_SPACE",
        pygame.K_EXCLAIM: "K_EXCLAIM",
        pygame.K_QUOTEDBL: "K_QUOTEDBL",
        pygame.K_HASH: "K_HASH",
        pygame.K_DOLLAR: "K_DOLLAR",
        pygame.K_AMPERSAND: "K_AMPERSAND",
        pygame.K_QUOTE: "K_QUOTE",
        pygame.K_LEFTPAREN: "K_LEFTPAREN",
        pygame.K_RIGHTPAREN: "K_RIGHTPAREN",
        pygame.K_ASTERISK: "K_ASTERISK",
        pygame.K_PLUS: "K_PLUS",
        pygame.K_COMMA: "K_COMMA",
        pygame.K_MINUS: "K_MINUS",
        pygame.K_PERIOD: "K_PERIOD",
        pygame.K_SLASH: "K_SLASH",
        pygame.K_0: "K_0",
        pygame.K_1: "K_1",
        pygame.K_2: "K_2",
        pygame.K_3: "K_3",
        pygame.K_4: "K_4",
        pygame.K_5: "K_5",
        pygame.K_6: "K_6",
        pygame.K_7: "K_7",
        pygame.K_8: "K_8",
        pygame.K_9: "K_9",
        pygame.K_COLON: "K_COLON",
        pygame.K_SEMICOLON: "K_SEMICOLON",
        pygame.K_LESS: "K_LESS",
        pygame.K_EQUALS: "K_EQUALS",
        pygame.K_GREATER: "K_GREATER",
        pygame.K_QUESTION: "K_QUESTION",
        pygame.K_AT: "K_AT",
        pygame.K_LEFTBRACKET: "K_LEFTBRACKET",
        pygame.K_BACKSLASH: "K_BACKSLASH",
        pygame.K_RIGHTBRACKET: "K_RIGHTBRACKET",
        pygame.K_CARET: "K_CARET",
        pygame.K_UNDERSCORE: "K_UNDERSCORE",
        pygame.K_BACKQUOTE: "K_BACKQUOTE",
        pygame.K_a: "K_a",
        pygame.K_b: "K_b",
        pygame.K_c: "K_c",
        pygame.K_d: "K_d",
        pygame.K_e: "K_e",
        pygame.K_f: "K_f",
        pygame.K_g: "K_g",
        pygame.K_h: "K_h",
        pygame.K_i: "K_i",
        pygame.K_j: "K_j",
        pygame.K_k: "K_k",
        pygame.K_l: "K_l",
        pygame.K_m: "K_m",
        pygame.K_n: "K_n",
        pygame.K_o: "K_o",
        pygame.K_p: "K_p",
        pygame.K_q: "K_q",
        pygame.K_r: "K_r",
        pygame.K_s: "K_s",
        pygame.K_t: "K_t",
        pygame.K_u: "K_u",
        pygame.K_v: "K_v",
        pygame.K_w: "K_w",
        pygame.K_x: "K_x",
        pygame.K_y: "K_y",
        pygame.K_z: "K_z",
        pygame.K_DELETE: "K_DELETE",
        pygame.K_KP0: "K_KP0",
        pygame.K_KP1: "K_KP1",
        pygame.K_KP2: "K_KP2",
        pygame.K_KP3: "K_KP3",
        pygame.K_KP4: "K_KP4",
        pygame.K_KP5: "K_KP5",
        pygame.K_KP6: "K_KP6",
        pygame.K_KP7: "K_KP7",
        pygame.K_KP8: "K_KP8",
        pygame.K_KP9: "K_KP9",
        pygame.K_KP_PERIOD: "K_KP_PERIOD",
        pygame.K_KP_DIVIDE: "K_KP_DIVIDE",
        pygame.K_KP_MULTIPLY: "K_KP_MULTIPLY",
        pygame.K_KP_MINUS: "K_KP_MINUS",
        pygame.K_KP_PLUS: "K_KP_PLUS",
        pygame.K_KP_ENTER: "K_KP_ENTER",
        pygame.K_KP_EQUALS: "K_KP_EQUALS",
        pygame.K_UP: "K_UP",
        pygame.K_DOWN: "K_DOWN",
        pygame.K_RIGHT: "K_RIGHT",
        pygame.K_LEFT: "K_LEFT",
        pygame.K_INSERT: "K_INSERT",
        pygame.K_HOME: "K_HOME",
        pygame.K_END: "K_END",
        pygame.K_PAGEUP: "K_PAGEUP",
        pygame.K_PAGEDOWN: "K_PAGEDOWN",
        pygame.K_F1: "K_F1",
        pygame.K_F2: "K_F2",
        pygame.K_F3: "K_F3",
        pygame.K_F4: "K_F4",
        pygame.K_F5: "K_F5",
        pygame.K_F6: "K_F6",
        pygame.K_F7: "K_F7",
        pygame.K_F8: "K_F8",
        pygame.K_F9: "K_F9",
        pygame.K_F10: "K_F10",
        pygame.K_F11: "K_F11",
        pygame.K_F12: "K_F12",
        pygame.K_F13: "K_F13",
        pygame.K_F14: "K_F14",
        pygame.K_F15: "K_F15",
        pygame.K_NUMLOCK: "K_NUMLOCK",
        pygame.K_CAPSLOCK: "K_CAPSLOCK",
        pygame.K_SCROLLOCK: "K_SCROLLOCK",
        pygame.K_RSHIFT: "K_RSHIFT",
        pygame.K_LSHIFT: "K_LSHIFT",
        pygame.K_RCTRL: "K_RCTRL",
        pygame.K_LCTRL: "K_LCTRL",
        pygame.K_RALT: "K_RALT",
        pygame.K_LALT: "K_LALT",
        pygame.K_RMETA: "K_RMETA",
        pygame.K_LMETA: "K_LMETA",
        pygame.K_LSUPER: "K_LSUPER",
        pygame.K_RSUPER: "K_RSUPER",
        pygame.K_MODE: "K_MODE",
        pygame.K_HELP: "K_HELP",
        pygame.K_PRINT: "K_PRINT",
        pygame.K_SYSREQ: "K_SYSREQ",
        pygame.K_BREAK: "K_BREAK",
        pygame.K_MENU: "K_MENU",
        pygame.K_POWER: "K_POWER",
        pygame.K_EURO: "K_EURO"}

    return mapper.get(val, None)






import pygame
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
from coopstructs.vectors import Vector2, IVector, Vector3
import numpy as np
from typing import List
import logging
from coopgame.pyLabel import TextAlignmentType
from coopbugger.monitoredclass import MonitoredClass

def mouse_pos_as_vector() -> Vector2:
    """ Get the global coords of the mouse position and convert them to a Vector2 object"""
    pos = pygame.mouse.get_pos()
    return Vector2(pos[0], pos[1])

def draw_box(surface: pygame.Surface, rect: Rectangle, color: Color, width: int = 0):
    pygame.draw.rect(surface, color.value, (rect.x, rect.y, rect.width, rect.height), width)


def draw_polygon(surface: pygame.Surface,
                 points,
                 color: Color,
                 width: int = 0,
                 alpha: int = 255):
    if type(points[0]) in [IVector, Vector2]:
        points = [point.as_tuple() for point in points]

    minx = min(point[0] for point in points)
    maxx = max(point[0] for point in points)
    miny = min(point[1] for point in points)
    maxy = max(point[1] for point in points)

    points = [(point[0] - minx, point[1] - miny) for point in points]

    # s = pygame.Surface((int(maxx - minx), int(maxy - miny)))
    # s.set_alpha(alpha)
    # s.set_colorkey(Color.HOT_PINK.value)
    # s.fill(Color.HOT_PINK.value)
    # pygame.draw.polygon(s, color.value, points, width)

    s = pygame.Surface((int(maxx - minx), int(maxy - miny)), pygame.SRCALPHA)
    c_with_alpha = (color.value[0], color.value[1], color.value[2], alpha)
    pygame.draw.polygon(s, c_with_alpha, points, width)

    surface.blit(s, (minx, miny))


def game_area_coords_from_parent_coords(parent_coords: Vector2, game_area_surface_rectangle: Rectangle) -> Vector2:
    """Converts Global Coords into coords on the game area"""
    return Vector2(parent_coords.x - game_area_surface_rectangle.x, parent_coords.y - game_area_surface_rectangle.y)


def scaled_points_of_a_rect(rect, grid_pos: Vector2, draw_scale_matrix=None, margin: int = 0):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    ''' get the rectangle object representing the grid position that was input'''
    rect = Rectangle(x=(margin + rect.width) * grid_pos.x + margin
                     , y=(margin + rect.height) * grid_pos.y + margin
                     , height=rect.height
                     , width=rect.width)

    '''Convert the rectangle to a list of points at the 4 corners'''
    points = [(x[0], x[1], 0, 1) for x in rect.points_tuple()]

    '''Multiply the points by the transform matrix for drawing'''
    transformed_points = draw_scale_matrix.dot(np.transpose(points))  # Transpose the points to appropriately mutiply

    '''return the x and y position for all points on the rectangle'''
    return np.transpose(transformed_points)[:,
           :3]  # Re-Transpose the points back to remain in a "list of points" format


def scaled_points(points: List[Vector2], draw_scale_matrix=None):
    scaled_array = scaled_points_of_points(points, draw_scale_matrix)

    return [Vector2(point[0], point[1]) for point in scaled_array]


def scaled_points_of_points(points: List[Vector2], draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    if len(points) == 0:
        return np.array([])
    '''Convert the point to a 4-dim point for multiplication'''
    normal_points = [(point.x, point.y, 0, 1) for point in points]

    '''Multiply the points by the transform matrix for drawing'''
    transformed_points = draw_scale_matrix.dot(
        np.transpose(normal_points))  # Transpose the points to appropriately mutiply

    return np.transpose(transformed_points)[:,
           :3]  # Re-Transpose the points back to remain in a "list of points" format


def viewport_point_on_plane(viewport_point: Vector2, game_area_rect, draw_scale_matrix=None, margin: int = 1):
    points_on_plane = scaled_points_of_a_rect(game_area_rect, Vector2(0, 0), draw_scale_matrix, margin=margin)
    point0 = points_on_plane[0]
    point1 = points_on_plane[1]
    point2 = points_on_plane[2]

    vec1 = point0 - point1
    vec2 = point2 - point1

    normal = np.cross(vec1, vec2)
    a = normal[0]
    b = normal[1]
    c = normal[2]
    d = a * point0[0] + b * point0[1] + c * point0[2]

    z_val = (d - a * viewport_point.x - b * viewport_point.y) / c

    return (viewport_point.x, viewport_point.y, z_val)


def scaled_points_to_normal_points(points: List[IVector], draw_scale_matrix=None):
    translated_points = [(point.x, point.y, 0, 1) for point in points]
    normal_array = scaled_array_to_normal_array(scaled_array=translated_points, draw_scale_matrix=draw_scale_matrix)
    normal_points = [Vector2(point[0], point[1]) for point in normal_array]

    return normal_points


def scaled_array_to_normal_array(scaled_array, draw_scale_matrix=None):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    draw_scale_matrix_inv = np.linalg.inv(draw_scale_matrix)
    normal_points = draw_scale_matrix_inv.dot(np.transpose(scaled_array))

    normal_points = np.reshape(normal_points, (-1, 2))

    return normal_points


def normal_points_to_scaled_points(points: List[Vector2], draw_scale_matrix=None):
    return scaled_points_of_points(points, draw_scale_matrix)


def mouse_in_plane_point(game_area_rect: Rectangle, draw_scale_matrix=None):
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(),
                                                                game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = viewport_point_on_plane(mouse_game_area_coord, game_area_rect, draw_scale_matrix, margin=1)

    return mouse_plane_point


def normal_mouse_position(game_area_rect: Rectangle, draw_scale_matrix=None) -> Vector2:
    # test = draw_scale_matrix_inv.dot(draw_scale_matrix)
    """Gets the mouse position and converts it to a grid position"""
    # mouse_game_area_coord = game_area_coords_from_parent_coords(parent_coords=mouse_pos_as_vector(), game_area_surface_rectangle=game_area_rect)
    mouse_plane_point = mouse_in_plane_point(game_area_rect, draw_scale_matrix)
    points = np.array([mouse_plane_point[0], mouse_plane_point[1], mouse_plane_point[2], 1])

    normal_array = scaled_array_to_normal_array(points, draw_scale_matrix)

    return Vector2(normal_array[0][0], normal_array[0][1])


def draw_text(text: str,
              hud: pygame.Surface,
              font: pygame.font.Font = None,
              offset_rect: Rectangle = None,
              color: Color = None,
              alignment: TextAlignmentType = None):
    if font is None:
        font = pygame.font.Font(None, 30)

    if offset_rect is None:
        offset_rect = Rectangle(0, hud.get_height() - 50, 20, hud.get_width())

    if color is None:
        color = Color.BLUE

    if alignment is None:
        alignment = TextAlignmentType.TOPLEFT

    rendered_txt = font.render(text, True, color.value)

    try:
        rect = rendered_txt.get_rect()
        align_coords = alignment_coords_for_type_rect(alignment, offset_rect).as_tuple()
        setattr(rect, alignment.value, align_coords)
        hud.blit(rendered_txt, rect)
    except Exception as e:
        logging.error(f"{e}")

def alignment_coords_for_type_rect(alignment: TextAlignmentType, rect: Rectangle) -> Vector2:
    switch = {
        TextAlignmentType.TOPRIGHT: lambda: rect.top_right,
        TextAlignmentType.TOPLEFT: lambda: rect.top_left
    }

    return switch.get(alignment)()

def draw_monitoredclass_stats(monitoredClass: MonitoredClass,
                              surface: pygame.Surface,
                              total_game_time_sec: float,
                              font: pygame.font.SysFont = None,
                              offset_rectangle: Rectangle = None,
                              g_offset: int = 0,
                              font_color: Color = None):

    if font is None:
        font = pygame.font.SysFont(None, 18)

    if font_color is None:
        font_color = Color.LEMON_CHIFFON

    tracked_time = [(key, val) for key, val in monitoredClass.tracked_time.items()]
    tracked_time.sort(key=lambda x: x[1], reverse=True)

    txt_lmbda = lambda key, val: f"{key}: {round(val, 2)} sec ({round(val / total_game_time_sec * 100, 1)}%)"

    if offset_rectangle is None:
        max_wid = max(font.size(txt_lmbda(key, val))[0] for key, val in tracked_time)
        offset_rectangle = Rectangle(0, 0, font.get_height() + 3, max_wid)

    y_off = offset_rectangle.y + g_offset
    offset_rectangle.y = y_off
    draw_text(f"RunTime Stats for {type(monitoredClass).__name__}", surface, color=font_color,
                   offset_rect=offset_rectangle, font=font)
    y_off += font.get_height() + 3
    for key, val in tracked_time:
        offset_rectangle.y = y_off
        draw_text(txt_lmbda(key, val), surface, color=font_color,
                       offset_rect=offset_rectangle, font=font, alignment=TextAlignmentType.TOPRIGHT)
        y_off += font.get_height() + 3

    return y_off

def draw_fps(hud: pygame.Surface,
             fps: float,
             font: pygame.font.Font = None,
             offset_rect: Rectangle = None,
             color: Color = None,
             alignment: TextAlignmentType = None):
    txt = f"FPS: {int(fps)}"
    draw_text(txt, hud=hud, font=font, offset_rect=offset_rect, color=color, alignment=alignment)

def draw_mouse_coord(hud: pygame.Surface, font: pygame.font.Font = None, offset_rect: Rectangle = None,
                     color: Color = None, alignment: TextAlignmentType = None):
    mouse_pos = mouse_pos_as_vector()
    txt = f"M:<{int(mouse_pos.x)}, {int(mouse_pos.y)}>"
    draw_text(txt, hud=hud, font=font, offset_rect=offset_rect, color=color, alignment=alignment)

def draw_game_time(hud: pygame.Surface,
                   game_time_s: float,
                   font: pygame.font.Font = None,
                   offset_rect: Rectangle = None,
                   color: Color = None,
                   alignment: TextAlignmentType = None):
    hrs = int(game_time_s / 3600)
    min = int((game_time_s - hrs * 3600) / 60)
    s = round(game_time_s - hrs * 3600 - min * 60, 2)
    txt = f"GameTime: {hrs} hrs, {min} min, {s} sec"
    draw_text(txt, hud=hud, font=font, offset_rect=offset_rect, color=color, alignment=alignment)


import logging

import pygame
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
import functools
from coopgame.pygbutton import PygButton
from coopbugger.monitoredclass import MonitoredClass
from typing import Callable, List, Dict, Optional
import coopgame.pygamehelpers as help
from coopgame.logger import logger
from coopgame.sprites import AnimatedSprite
from coopstructs.geometry import Rectangle as cRect
from coopgame.monitoredClassLogger import MonitoredClassLogger
from cooptools.coopEnum import CoopEnum
from enum import auto
import inspect
from coopgame.pygame_k_constant_names import pygame_key_mapper
from coopgame.fpsTracker import FpsTracker
from coopstructs.toggles import BooleanToggleable
from cooptools.dataRefresher import DataRefresher
from cooptools.timedDecay import Timer
from coopgame.pyLabel import PyLabel, TextAlignmentType


def try_handler(func):
    @functools.wraps(func)
    def wrapper_handler(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except NotImplementedError as e:
            error = f"Inherited class should implement logic for {func.__name__}"
            logger.debug(error)
        except Exception as e:
            logger.exception(e)

    return wrapper_handler


class BuiltInSurfaceType(CoopEnum):
    BACKGROUND = auto()
    FOREGROUND = auto()
    HELP = auto()
    DEBUGGER = auto()


class GameTemplate(MonitoredClass):

    def __init__(self,
                 fullscreen: bool = False,
                 screen_width: int = 1000,
                 screen_height: int = 500,
                 max_fps: int = 120,
                 debug_mode=False):
        super().__init__()
        self.init_screen_width = screen_width
        self.init_screen_height = screen_height

        self.fullscreen = fullscreen
        self.screen = None
        self._init_screen(self.fullscreen, self.screen_width, self.screen_height)

        self.fps_tracker = FpsTracker(max_fps=max_fps)

        self.running = False
        self.debug_mode_toggle = BooleanToggleable(default=debug_mode)
        self.debug_surface_update_timer = Timer(200, True)

        self.buttons = {}
        self._key_handlers = {
            (pygame.K_ESCAPE,): (self.quit, False),
            (pygame.K_F12,): (self.toggle_fullscreen, False),
            (pygame.K_F11,): (self.debug_mode_toggle.toggle, False)

        }
        self.register_keys()

        self.sprites = {}

        self._monitored_class_logger = MonitoredClassLogger(monitored_classes=[self])

        # build in surfaces
        self.built_in_surfaces: Dict[BuiltInSurfaceType, Optional[pygame.Surface]] = {
            BuiltInSurfaceType.BACKGROUND: None,
            BuiltInSurfaceType.FOREGROUND: None,
            BuiltInSurfaceType.HELP: None,
            BuiltInSurfaceType.DEBUGGER: None
        }

        self.fps_tracker.set_start()
        self.log_stats_refresher = DataRefresher('log_stats',
                                                 refresh_callback=lambda: logger.info(
                                                     f"FPS: {int(self.fps_tracker.fps) if self.fps_tracker.fps else None}"),
                                                 refresh_interval_ms=1000)
        pygame.init()

    def quit(self):
        self.running = False

    def toggle_fullscreen(self):
        self.fullscreen = not self.fullscreen
        self._init_screen(self.fullscreen, self.screen_width, self.screen_height)

    def _init_screen(self, fullscreen=None, screen_width=None, screen_height=None):
        if fullscreen:
            self.screen = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)
        elif screen_width and screen_height:
            self.screen = pygame.display.set_mode((self.init_screen_width, self.init_screen_height))
        else:
            self.screen = pygame.display.set_mode((0, 0))

    def register_button(self, id, text, callback, postion_rect):
        self.buttons[id] = PygButton(postion_rect, caption=text, callback=callback)

    def main(self, debug: bool = False):
        self.debug_mode_toggle.set_value(debug)

        self.init_builtin_surfaces()
        self.initialize_game()

        self.running = True
        ii = 0
        while self.running:
            self._update()
            self._draw(frames=ii)
            self.fps_tracker.clock_tick()
            ii += 1

        pygame.quit()

    def _update(self):
        """:return
            Update environment based on time delta and any input
        """

        ''' Calculate the ticks between update calls so that the update functions can handle correct time deltas '''
        delta_time_ms = self.fps_tracker.update()

        '''Log Stats'''
        self._monitored_class_logger.check_and_log(delta_time_ms)
        self.log_stats_refresher.check_and_refresh()

        '''Update Model'''
        self._model_updater(delta_time_ms)

        '''Update Sprites'''
        self.sprite_updater(delta_time_ms)

        '''Animate Sprites'''
        self._sprite_animator(delta_time_ms)

        '''Handle Events'''
        self._handle_events()

    def _handle_events(self):
        """:return
            handle all of the registered events that have been captured since last iteration
        """

        '''Get next event'''
        for event in pygame.event.get():

            '''Get pressed keys'''
            pressed_keys = pygame.key.get_pressed()

            '''Check and handle button press'''
            self.handle_buttons(event)

            '''Debug Printer'''
            if event.type not in (0, 1, 4, 6):
                logger.debug(f"Pygame EventType: {event.type}")

            '''Event Type Switch'''
            if event.type == pygame.QUIT:
                self.quit()
            elif event.type == pygame.KEYDOWN:
                self._handle_key_pressed(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
                logger.info(f"Left Mouse Down")
                self.handle_left_mouse_down(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONUP and event.button == 1:
                logger.info(f"Left Mouse Up")
                self.handle_left_mouse_up(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 3:
                logger.info(f"Right Mouse Down")
                self.handle_right_mouse_down(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONUP and event.button == 3:
                logger.info(f"Right Mouse Up")
                self.handle_right_mouse_up(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 4:
                logger.info(f"Scroll Up")
                self.handle_mouse_scroll_up(pressed_keys)
            elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 5:
                logger.info(f"Scroll Down")
                self.handle_mouse_scroll_down(pressed_keys)
            elif event.type == pygame.WINDOWSIZECHANGED:
                logger.info(f"Window Resized [{self.screen.get_width()}x{self.screen.get_height()}]")
                self.on_resize()
            else:
                logger.debug(f"Unhandled event: {event.type}")

        '''Get pressed keys'''
        pressed_keys = pygame.key.get_pressed()

        '''Handle Held Keys'''
        if pressed_keys is not None:
            self._handle_held_keys(pressed_keys)

        '''Handle hover over'''
        self._handle_hover_over()

    def register_action_to_keys(self, keys_tuple, func: Callable, react_while_holding: bool = False):
        """
            Takes a tuple of keys (integers) and maps that key combo to a callable function

            :param keys_tuple a list of keys (integers) that will be mapped to the input callable. Note that a single
            value Tuple is input as ([key],) *note the comma
            :param func a callable that is mapped to the key combo
        """
        self._key_handlers[keys_tuple] = (func, react_while_holding)

    @MonitoredClass.timer
    @try_handler
    def handle_buttons(self, event):
        for id, button in self.buttons:
            if 'click' in button.handleEvent(event):
                button.callback()

    @MonitoredClass.timer
    @try_handler
    def _handle_key_pressed(self, pressed_keys):
        buttons = [pygame.key.name(k) for k, v in enumerate(pressed_keys) if v]
        logger.info(f"Keys pressed: {buttons}")
        for mapped_keys, reaction in self._key_handlers.items():
            func = reaction[0]
            if (all(pressed_keys[key] for key in mapped_keys)):
                func()

    @MonitoredClass.timer
    @try_handler
    def _handle_held_keys(self, pressed_keys):
        for mapped_keys, reaction in self._key_handlers.items():
            func = reaction[0]
            react_while_holding = reaction[1]
            if (all(pressed_keys[key] for key in mapped_keys)) and react_while_holding:
                func()

    @MonitoredClass.timer
    @try_handler
    def _handle_hover_over(self):
        mouse_pos_as_vector = help.mouse_pos_as_vector()
        if mouse_pos_as_vector:
            self.handle_hover_over(mouse_pos_as_vector)

    @MonitoredClass.timer
    @try_handler
    def _model_updater(self, delta_time_ms: int):
        return self.model_updater(delta_time_ms)

    @try_handler
    def initialize_game(self):
        raise NotImplementedError()

    @try_handler
    def handle_left_mouse_down(self, pressed_keys):
        raise NotImplementedError()

    @try_handler
    def handle_left_mouse_up(self, pressed_keys):
        raise NotImplementedError()

    @try_handler
    def handle_right_mouse_down(self, pressed_keys):
        raise NotImplementedError()

    @try_handler
    def handle_right_mouse_up(self, pressed_keys):
        raise NotImplementedError()

    @try_handler
    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        raise NotImplementedError()

    @try_handler
    def handle_mouse_scroll_up(self, pressed_keys):
        raise NotImplementedError()

    @try_handler
    def handle_mouse_scroll_down(self, pressed_keys):
        raise NotImplementedError()

    @try_handler
    def draw(self, frames: int, debug_mode: bool = False):
        raise NotImplementedError()

    @try_handler
    def model_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @try_handler
    def sprite_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @try_handler
    def on_resize(self):
        raise NotImplementedError()

    @try_handler
    def register_keys(self):
        raise NotImplementedError()

    @try_handler
    def update_built_in_surfaces(self, surface_types: List[BuiltInSurfaceType]):
        raise NotImplementedError()

    @MonitoredClass.timer
    def _draw(self, frames: int):
        self.screen.fill(Color.BLACK.value)
        self.draw(frames, self.debug_mode_toggle.value)

        '''Draw Sprites'''
        self.sprite_drawer(self.screen)

        if self.debug_mode_toggle.value:
            self._check_update_debugger_surface()
            self.screen.blit(self.built_in_surfaces[BuiltInSurfaceType.DEBUGGER], (0, 0))

        # Update the display
        pygame.display.flip()

    @MonitoredClass.timer
    def _check_update_debugger_surface(self):
        if self.debug_surface_update_timer.finished:
            self._update_debugger_surface()
            self.debug_surface_update_timer.reset()

    def _update_debugger_surface(self):
        deb_surf = self.built_in_surfaces[BuiltInSurfaceType.DEBUGGER]
        if deb_surf is None:
            return

        deb_surf.fill(Color.BLACK.value)
        help.draw_text(f"Sprite Count: {len(self.sprites)}", deb_surf,
                       offset_rect=cRect(self.screen.get_width() - 400, self.screen.get_height() - 110, 20, 400),
                       alignment=TextAlignmentType.TOPRIGHT)
        help.draw_fps(deb_surf,
                      fps=self.fps_tracker.fps,
                      offset_rect=cRect(self.screen.get_width() - 400, self.screen.get_height() - 80, 20, 400),
                      alignment=TextAlignmentType.TOPRIGHT
                      )
        help.draw_mouse_coord(deb_surf,
                              offset_rect=cRect(self.screen.get_width() - 400, self.screen.get_height() - 20, 20, 400),
                              alignment=TextAlignmentType.TOPRIGHT)
        help.draw_game_time(deb_surf,
                            game_time_s=self.fps_tracker.ticks / 1000,
                            offset_rect=cRect(self.screen.get_width() - 400, self.screen.get_height() - 50, 20, 400),
                            alignment=TextAlignmentType.TOPRIGHT)

        self.draw_monitoredclass_stats(surface=deb_surf)

    def draw_monitoredclass_stats(self,
                                  surface: pygame.Surface):
        offset = 0
        for mc in self._monitored_class_logger.monitored_classes:
            offset = help.draw_monitoredclass_stats(monitoredClass=mc,
                                                    surface=surface,
                                                    g_offset=offset,
                                                    total_game_time_sec=self.fps_tracker.ticks / 1000)

    def _sprite_animator(self,
                         delta_time_ms: int):

        for name, sprite in self.sprites.items():
            if type(sprite) == AnimatedSprite:
                sprite.animate(delta_time_ms)

    def sprite_drawer(self,
                      surface: pygame.surface):

        sprites = list(self.sprites.values())
        sprites.sort(key=lambda x: x.bottom_center_pos.y)
        for sprite in sprites:
            sprite.blit(surface, display_handle=self.debug_mode_toggle.value, display_rect=self.debug_mode_toggle.value)

    def register_monitored_classes(self, new_classes: List[MonitoredClass]):
        self._monitored_class_logger.register_classes(new_classes)

    def init_builtin_surfaces(self, types: List[BuiltInSurfaceType] = None, colorkey: Color = None):
        if colorkey is None: colorkey = Color.BLACK
        if types is None: types = [e for e in BuiltInSurfaceType]

        for type in types:
            self.built_in_surfaces[type] = pygame.Surface(self.screen.get_size()).convert()
            self.built_in_surfaces[type].set_colorkey(colorkey.value)

    def _update_help_surface(self):
        text = self._help_txt()

        self.built_in_surfaces[BuiltInSurfaceType.HELP].fill(Color.WHEAT.value)

        font = pygame.font.Font(None, 20)
        fontsize = font.get_height()
        offSet = 0

        for idx, line in enumerate(text):
            help.draw_text(line,
                           self.built_in_surfaces[BuiltInSurfaceType.HELP],
                           font=font,
                           offset_rect=Rectangle(0,
                                                 idx * fontsize + offSet,
                                                 fontsize + 3,
                                                 self.built_in_surfaces[BuiltInSurfaceType.HELP].get_width()))

    def _help_txt(self):

        txt = []
        txt.append("HELP")
        txt.append("\n--------")
        txt.append("\n")
        for k, v in self._key_handlers.items():
            if len(v) > 2:
                callback_txt = v[2]
            elif v[0].__name__ == "<lambda>":
                callback_txt = inspect.getsource(v[0])
                callback_txt = callback_txt.split("lambda:", 1)[1].split('(', 1)[0].replace("self.", "")
            else:
                callback_txt = v[0].__name__
            k_tup = [pygame_key_mapper(x) for x in k]
            txt.append(f"{k_tup} -- {callback_txt}")

        return txt

    @property
    def screen_width(self):
        return self.screen.get_width() if self.screen else self.init_screen_width

    @property
    def screen_height(self):
        return self.screen.get_height() if self.screen else self.init_screen_height

    @property
    def game_area_rectangle(self):
        return Rectangle(0, 0, self.screen.get_height(), self.screen.get_width())

if __name__ == "__main__":
    gt = GameTemplate()
    gt.main(debug=True)


from coopbugger.monitoredclass import MonitoredClass
from typing import List
from coopgame.logger import logger


class MonitoredClassLogger():

    def __init__(self,
                 log_interval_ms: int = 10000,
                 monitored_classes: List[MonitoredClass] = None):
        self.time_since_report_stats = 0
        self.log_interval_ms = log_interval_ms
        self.monitored_classes = monitored_classes if monitored_classes else []

    def check_and_log(self,
                      delta_time_ms: int):
        self.time_since_report_stats += delta_time_ms
        if self.time_since_report_stats > self.log_interval_ms:
            for item in self.monitored_classes:
                print_dict = {method: round(time, 3) for method, time in item.tracked_time.items()}
                logger.info(f"{item} -- {print_dict}")

            self.time_since_report_stats = 0

    def register_classes(self, new: List[MonitoredClass]):
        self.monitored_classes += new


import time

import pygame


class FpsTracker:
    def __init__(self, max_fps):
        self.ticks = 0
        self.frame_times = []
        self.fps = None
        self.max_fps = max_fps

        self.clock = pygame.time.Clock()
        self.game_start = None

    def calculate_fps(self, ticks_last_frame: int):
        if len(self.frame_times) > 20:
            self.frame_times.pop(0)

        self.frame_times.append(ticks_last_frame)

        avg_sec_per_frame = sum(self.frame_times) / len(self.frame_times) / 1000.0
        self.fps = 1 / avg_sec_per_frame if avg_sec_per_frame > 0 else 0

    def clock_tick(self):
        self.clock.tick(self.max_fps)

    def set_start(self):
        self.game_start = time.perf_counter()

    @property
    def run_time(self):
        return (time.perf_counter() - self.game_start) if self.game_start else None

    def update(self) -> int:
        t = pygame.time.get_ticks()
        delta_time_ms = (t - self.ticks)
        self.ticks = t
        self.calculate_fps(delta_time_ms)

        return delta_time_ms



import pygame
from coopgame.colors import Color
from typing import List, Dict, Callable
from coopstructs.curves import Arc, CubicBezier, LineCurve, Curve, Orientation, CatmullRom, CircularArc
from coopstructs.geometry import Line, PolygonRegion
import numpy as np
from descartes import PolygonPatch
from shapely.geometry import LineString
import coopgame.pygamehelpers as help


class CurveDrawer:

    def __init__(self):
        pass

    @classmethod
    def draw_curves(cls,
                    curves: Dict[Curve, Color],
                    surface: pygame.Surface,
                    control_point_color: Color = Color.RED,
                    control_point_size: int = None,
                    control_line_color: Color = Color.GREY,
                    draw_scale_matrix=None,
                    buffer: int = None,
                    buffer_color: Color = None):
        for curve, color in curves.items():
            if type(curve) == CubicBezier:
                cls.draw_bezier(surface,
                                 curve,
                                 color,
                                 control_point_color=control_point_color,
                                 control_point_size=control_point_size,
                                 control_line_color=control_line_color,
                                 draw_scale_matrix=draw_scale_matrix,
                                 buffer=buffer,
                                 buffer_color=buffer_color)
            elif type(curve) == LineCurve:
                cls.draw_line(surface,
                               curve,
                               color,
                               control_point_color=control_point_color,
                               control_point_size=control_point_size,
                               draw_scale_matrix=draw_scale_matrix,
                               buffer=buffer,
                               buffer_color=buffer_color)
            elif type(curve) == Arc:
                cls.draw_arc(surface,
                              curve,
                              color,
                              control_point_color=control_point_color,
                              control_point_size=control_point_size,
                              draw_scale_matrix=draw_scale_matrix,
                              buffer=buffer,
                              buffer_color=buffer_color)
            elif type(curve) == CatmullRom:
                cls.draw_catmullrom(surface,
                                     curve,
                                     color,
                                     control_point_color=control_point_color,
                                     control_point_size=control_point_size,
                                     draw_scale_matrix=draw_scale_matrix,
                                     buffer=buffer,
                                     buffer_color=buffer_color)
            elif type(curve) == CircularArc:
                cls.draw_circulararc(surface,
                                      curve,
                                      color,
                                      control_point_color=control_point_color,
                                      control_point_size=control_point_size,
                                      draw_scale_matrix=draw_scale_matrix,
                                      buffer=buffer,
                                      buffer_color=buffer_color)
            else:
                raise NotImplementedError(f"Unahanlded curve type {curve}")

    @classmethod
    def draw_bezier(cls,
                    screen,
                    curve: CubicBezier,
                    curve_color: Color,
                    control_point_color: Color = Color.RED,
                    control_point_size: int = None,
                    control_line_color: Color = Color.GREY,
                    draw_scale_matrix=None,
                    buffer: int = None,
                    buffer_color: Color = None):

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        arrays_of_points = []
        for point in curve.control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw bezier curve
        b_points = curve.compute_bezier_points([(translated_points[ii][0], translated_points[ii][1])
                                                for ii in range(len(translated_points))])

        if buffer is not None:
            line = LineString(b_points)

            dilated = line.buffer(buffer)
            poly = PolygonRegion.from_shapely_polygon(dilated)
            buffer_color = Color.GREEN if buffer_color is None else buffer_color
            help.draw_polygon(screen, [x.as_tuple() for x in poly.boundary_points], buffer_color)

        # Draw control points
        if control_point_size is None:
            control_point_size = 4

        if control_point_color:
            for ii in range(len(translated_points)):
                pygame.draw.circle(screen, control_point_color.value,
                                   (int(translated_points[ii][0]), int(translated_points[ii][1])), control_point_size)
            # for p in curve.control_points:
            #     pygame.draw.circle(screen, control_point_color.value, (int(p.x), int(p.y)), 4)

        # Draw control "lines"
        if control_line_color:
            pygame.draw.lines(screen,
                              control_line_color.value,
                              False,
                              [(translated_points[ii][0], translated_points[ii][1])
                               for ii in range(len(translated_points))])
            # pygame.draw.lines(screen, control_line_color.value, False, [(x.x, x.y) for x in curve.control_points])

        # b_points = curve.compute_bezier_points([(x.x, x.y) for x in curve.control_points])
        if b_points is not None and len(b_points) > 1:
            pygame.draw.lines(screen, curve_color.value, False, b_points, 2)

    @classmethod
    def draw_line(cls,
                  surface: pygame.Surface,
                  line: Line,
                  color: Color,
                  control_point_color: Color = Color.RED,
                  control_point_size: int = None,
                  draw_scale_matrix=None,
                  buffer: int = None,
                  buffer_color: Color = None):
        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert Origin/Destination into np.arrays
        origin = np.array([line.origin.x, line.origin.y, 0, 1])
        destination = np.array([line.destination.x, line.destination.y, 0, 1])

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array([origin, destination]))).transpose()

        # Draw the lines
        pygame.draw.line(surface, color.value, (translated_points[0][0], translated_points[0][1]),
                         (translated_points[1][0], translated_points[1][1]))

        # Draw the points
        if control_point_size is None:
            control_point_size = 4

        if control_point_color:
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[0][0]), int(translated_points[0][1])), control_point_size)
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[1][0]), int(translated_points[1][1])), control_point_size)

    @classmethod
    def draw_arc(cls,
                 surface: pygame.Surface,
                 arc: Arc,
                 color: Color,
                 control_point_color: Color = Color.RED,
                 control_point_size: int = None,
                 draw_scale_matrix=None,
                 buffer: int = None,
                 buffer_color: Color = None):

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        control_points = arc.compute_curve_points()
        arrays_of_points = []
        for point in control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw Lines
        pygame.draw.lines(surface, color.value, False, [(translated_points[ii][0], translated_points[ii][1])
                                                        for ii in range(len(translated_points))], 2)

        # Draw the points
        if control_point_size is None:
            control_point_size = 4

        if control_point_color:
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[0][0]), int(translated_points[0][1])), control_point_size)
            pygame.draw.circle(surface, control_point_color.value,
                               (int(translated_points[-1][0]), int(translated_points[-1][1])), control_point_size)

    @classmethod
    def draw_catmullrom(cls,
                        screen,
                        curve: CatmullRom,
                        curve_color: Color,
                        control_point_color: Color = Color.RED,
                        control_point_size: int = None,
                        control_line_color: Color = Color.GREY,
                        draw_scale_matrix=None,
                        buffer: int = None,
                        buffer_color: Color = None):

        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        arrays_of_points = []
        for point in curve.control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw control points
        if control_point_size is None:
            control_point_size = 4

        if control_point_color:
            for ii in range(len(translated_points)):
                pygame.draw.circle(screen, control_point_color.value,
                                   (int(translated_points[ii][0]), int(translated_points[ii][1])), control_point_size)

        # Draw control "lines"
        if control_line_color:
            pygame.draw.lines(screen,
                              control_line_color.value,
                              False,
                              [(translated_points[ii][0], translated_points[ii][1])
                               for ii in range(len(translated_points))])

        # Draw CatmullRom curve
        b_points = curve.compute_catmull_points([(translated_points[ii][0], translated_points[ii][1])
                                                 for ii in range(len(translated_points))])
        # b_points = curve.compute_catmull_points(curve.control_points)

        if b_points is not None and len(b_points) > 1:
            pygame.draw.lines(screen, curve_color.value, False, [(point.x, point.y) for point in b_points], 2)

    @classmethod
    def draw_circulararc(cls,
                         screen,
                         curve: CircularArc,
                         curve_color: Color,
                         control_point_color: Color = Color.RED,
                         control_point_size: int = None,
                         control_line_color: Color = Color.GREY,
                         draw_scale_matrix=None,
                         buffer: int = None,
                         buffer_color: Color = None):
        if draw_scale_matrix is None:
            draw_scale_matrix = np.identity(4)

        # Convert control points into np.arrays
        arrays_of_points = []
        for point in curve.control_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        # Draw control points
        if control_point_size is None:
            control_point_size = 4

        if control_point_color:
            for ii in range(len(translated_points)):
                pygame.draw.circle(screen, control_point_color.value,
                                   (int(translated_points[ii][0]), int(translated_points[ii][1])), control_point_size)

        # Draw control "lines"
        if control_line_color:
            pygame.draw.lines(screen,
                              control_line_color.value,
                              False,
                              [(translated_points[ii][0], translated_points[ii][1])
                               for ii in range(len(translated_points))])

        # Draw circulararc_points
        b_points = curve.compute_circulararc_points()

        # Convert
        arrays_of_points = []
        for point in b_points:
            arrays_of_points.append(np.array([point.x, point.y, 0, 1]))

        # Translate the points via the scaling matrix
        translated_points = draw_scale_matrix.dot(np.transpose(np.array(arrays_of_points))).transpose()

        if translated_points is not None and len(translated_points) > 1:
            pygame.draw.lines(screen, curve_color.value, False, [(point[0], point[1]) for point in translated_points],
                              2)

    @classmethod
    def draw_curve_builder(self):
        pass